@@ -45,6 +45,7 @@ module Database.Persist.TH
 
 import Prelude hiding ((++), take, concat, splitAt, exp)
 import Database.Persist
+import Database.Persist.Class (HasPersistBackend(..), BaseBackend)
 import Database.Persist.Sql (Migration, migrate, SqlBackend, PersistFieldSql)
 import Database.Persist.Quasi
 import Language.Haskell.TH.Lib (
@@ -80,6 +81,14 @@ import Web.HttpApiData (ToHttpApiData(..), FromHttpApiData(..), parseUrlPieceMay
 import GHC.Generics (Generic)
 import qualified Data.Text.Encoding as TE
 
+-- | This special-cases "type_" and strips out its underscore. When
+-- used for JSON serialization and deserialization, it works around
+-- <https://github.com/yesodweb/persistent/issues/412>
+unHaskellNameForJSON :: HaskellName -> Text
+unHaskellNameForJSON = fixTypeUnderscore . unHaskellName
+  where fixTypeUnderscore "type" = "type_"
+        fixTypeUnderscore name = name
+
 -- | Converts a quasi-quoted syntax into a list of entity definitions, to be
 -- used as input to the template haskell generation code (mkPersist).
 persistWith :: PersistSettings -> QuasiQuoter
@@ -244,7 +253,10 @@ setEmbedField entName allEntities field = field
               then EmbedRef em
               else if maybeNullable field
                      then SelfReference
-                     else error $ unpack $ unHaskellName entName `mappend` ": a self reference must be a Maybe"
+                     else case fieldType field of
+                       FTList _ -> SelfReference
+                       _ -> error $ unpack $ unHaskellName entName
+                           `mappend` ": a self reference must be a Maybe"
       existing@_   -> existing
   }
 
@@ -295,11 +307,12 @@ mkEntityDefSqlTypeExp emEntities entMap ent = EntityDefSqlTypeExp ent
 mkPersist :: MkPersistSettings -> [EntityDef] -> Q [Dec]
 mkPersist mps ents' = do
     x <- fmap mconcat $ mapM (persistFieldFromEntity mps) ents
-    y <- fmap mconcat $ mapM (mkEntity mps) ents
+    y <- fmap mconcat $ mapM (mkEntity entMap mps) ents
     z <- fmap mconcat $ mapM (mkJSON mps) ents
     return $ mconcat [x, y, z]
   where
     ents = map fixEntityDef ents'
+    entMap = M.fromList $ map (\ent -> (entityHaskell ent, ent)) ents
 
 -- | Implement special preprocessing on EntityDef as necessary for 'mkPersist'.
 -- For example, strip out any fields marked as MigrationOnly.
@@ -924,9 +937,9 @@ fromValues t funName conE fields = do
           mkPersistValue fieldName = [|mapLeft (fieldError fieldName) . fromPersistValue|]
 
 
-mkEntity :: MkPersistSettings -> EntityDef -> Q [Dec]
-mkEntity mps t = do
-    t' <- lift t
+mkEntity :: EntityMap -> MkPersistSettings -> EntityDef -> Q [Dec]
+mkEntity entMap mps t = do
+    t' <- liftAndFixKeys entMap t
     let nameT = unHaskellName entName
     let nameS = unpack nameT
     let clazz = ConT ''PersistEntity `AppT` genDataType
@@ -1206,7 +1219,7 @@ mkDeleteCascade mps defs = do
         return $
             instanceD
             [ mkClassP ''PersistQuery [backendT]
-            , mkEqualP (ConT ''PersistEntityBackend `AppT` entityT) backendT
+            , mkEqualP (ConT ''PersistEntityBackend `AppT` entityT) (ConT ''BaseBackend `AppT` backendT)
             ]
             (ConT ''DeleteCascade `AppT` entityT `AppT` backendT)
             [ FunD 'deleteCascade
@@ -1340,6 +1353,7 @@ mkMigrate fun allDefs = do
     defs = filter isMigrated allDefs
     isMigrated def = not $ "no-migrate" `elem` entityAttrs def
     typ = ConT ''Migration
+    entMap = M.fromList $ map (\ent -> (entityHaskell ent, ent)) allDefs
     body :: Q Exp
     body =
         case defs of
@@ -1347,17 +1361,46 @@ mkMigrate fun allDefs = do
             _  -> do
               defsName <- newName "defs"
               defsStmt <- do
-                defs' <- mapM lift defs
+                defs' <- mapM (liftAndFixKeys entMap) defs
                 let defsExp = ListE defs'
                 return $ LetS [ValD (VarP defsName) (NormalB defsExp) []]
               stmts <- mapM (toStmt $ VarE defsName) defs
               return (DoE $ defsStmt : stmts)
     toStmt :: Exp -> EntityDef -> Q Stmt
     toStmt defsExp ed = do
-        u <- lift ed
+        u <- liftAndFixKeys entMap ed
         m <- [|migrate|]
         return $ NoBindS $ m `AppE` defsExp `AppE` u
 
+liftAndFixKeys :: EntityMap -> EntityDef -> Q Exp
+liftAndFixKeys entMap EntityDef{..} =
+  [|EntityDef
+      entityHaskell
+      entityDB
+      entityId
+      entityAttrs
+      $(ListE <$> mapM (liftAndFixKey entMap) entityFields)
+      entityUniques
+      entityForeigns
+      entityDerives
+      entityExtra
+      entitySum
+   |]
+
+liftAndFixKey :: EntityMap -> FieldDef -> Q Exp
+liftAndFixKey entMap (FieldDef a b c sqlTyp e f fieldRef) =
+  [|FieldDef a b c $(sqlTyp') e f fieldRef'|]
+  where
+    (fieldRef', sqlTyp') = fromMaybe (fieldRef, lift sqlTyp) $
+      case fieldRef of
+        ForeignRef refName ft -> case M.lookup refName entMap of
+          Nothing -> Nothing
+          Just ent ->
+            case fieldReference $ entityId ent of
+              fr@(ForeignRef _Name ft) -> Just (fr, lift $ SqlTypeExp ft)
+              _ -> Nothing
+        _ -> Nothing
+
 instance Lift EntityDef where
     lift EntityDef{..} =
         [|EntityDef
@@ -1519,7 +1562,7 @@ mkJSON mps def = do
     obj <- newName "obj"
     mzeroE <- [|mzero|]
 
-    xs <- mapM (newName . unpack . unHaskellName . fieldHaskell)
+    xs <- mapM (newName . unpack . unHaskellNameForJSON . fieldHaskell)
         $ entityFields def
 
     let conName = mkName $ unpack $ unHaskellName $ entityHaskell def
@@ -1633,4 +1676,3 @@ instanceD = InstanceD
 --         let x = mkName "x"
 --          in normalClause [ConP (mkName constr) [VarP x]]
 --                    (VarE 'toPersistValue `AppE` VarE x)
-
